/******************************************************************************/
/* Files to Include (MP3_03)                                                  */
/******************************************************************************/

#if defined(__XC)
    #include <xc.h>        /* XC8 General Include File */
#elif defined(HI_TECH_C)
    #include <htc.h>       /* HiTech General Include File */
#elif defined(__18CXX)
    #include <p18cxxx.h>   /* C18 General Include File */
#endif

#if defined(__XC) || defined(HI_TECH_C)

#include <stdint.h>        /* For uint8_t definition */
#include <stdbool.h>       /* For true/false definition */
#include <string.h>
#include <stdlib.h>        /* rand */

#endif

#include "system.h"        /* System funct/params, like osc/peripheral config */
#include "user.h"          /* User funct/params, such as InitApp */
#include "ff.h"

#define MP3_BUFFER_SIZE     512           // 512=32*16
#define SEND_BUF_CNT_UNIT   200           // 512*200=102400=0.1MByte

#define SW1_PRESSED         0x81
#define SW2_PRESSED         0x82

#define PLAY_NEXT_MP3       0x91
#define PLAY_OTHER_DIR      0x92

#define MIN_VOLUME          0xD0
#define MAX_VOLUME          0x00

#define DIRLISTFILE         "/DIRLIST.TXT"      // dir /AD /ON /S /B > DIRLIST.TXT
#define HIST_SIZE           64
#define HIST_EEPROM_ADDRESS 16

/******************************************************************************/
/* User Global Variable Declaration                                           */
/******************************************************************************/

/* i.e. uint8_t <variable_name>; */
FATFS   Fatfs;

uint8_t     VOLUME;
uint16_t    ST_AMP, ST_FREQ, SB_AMP, SB_FREQ;   // bass/treble control values
static uint8_t     TREBLE_CONTROL;
static unsigned char	MP3DATA[MP3_BUFFER_SIZE];
static uint16_t    SrandSeed;
static uint16_t    Hist[HIST_SIZE];

uint8_t     STATE_SW1;
uint8_t     STATE_SW2;
uint16_t    CHATTERING_TMR;
uint16_t    SW_CONTINUE_COUNT;
uint8_t     LED_BLINK_MODE;

static uint8_t Level_0(char *path);
static uint8_t Level_1(TCHAR *parentdirpath, TCHAR *subdirname);
static uint8_t Level_2(TCHAR *parentdirpath, TCHAR *subdirname);
static uint8_t Level_3(TCHAR *parentdirpath, TCHAR *subdirname);
static uint8_t Level_4(TCHAR *parentdirpath, TCHAR *subdirname);
static uint8_t Make_SubDir_Path(TCHAR *outpath, uint8_t bufsize, TCHAR *parentdirpath, TCHAR *subdir);
static uint8_t count_mp3_dir(TCHAR *path, uint8_t *mp3count, uint8_t *dircount);
static uint8_t play_random();
static uint8_t check_hist(uint16_t no);
static void add_hist(uint16_t no);
static uint8_t play_subdirectory(TCHAR *path);
static uint8_t play_mp3(TCHAR *mp3file);
static uint8_t PauseAndSelectItem();
static uint8_t isMp3(char *fname);


/******************************************************************************/
/* Main Program (MP3_02)                                                      */
/******************************************************************************/
void main(void)
{
    FRESULT res;

    /* Configure the oscillator for the device */
    ConfigureOscillator();

    /* Initialize I/O and Peripherals for application */
    InitApp();
    VOLUME = MAX_VOLUME;
    ST_AMP = 0x00; ST_FREQ = 0x00; SB_AMP = 0x00; SB_FREQ = 0x00;
    TREBLE_CONTROL = 0;

    output_low(MP3_RS);
    wait_ms(100);
    output_high(MP3_RS);
    wait_ms(100);
    INTCONbits.TMR0IE = 1;      // Enables the TMR0 overflow interrupt

    STATE_SW2 = 0;
    STATE_SW1 = 0;
    CHATTERING_TMR = 0;
    LED = 0;

    res = f_mount(&Fatfs, _T("0:"), 0);
    if (res == FR_OK){
        uint8_t     sel;
        sel = PauseAndSelectItem();
        if (sel == SW1_PRESSED){
            if (f_stat(DIRLISTFILE, NULL) == FR_OK){    // directory list file
                // random play
                uint16_t    seed;
                seed = read_from_eeprom_uint16(0);
                if (seed == 0xFFFF) seed = 0;
                else seed++;
                seed ^= SrandSeed;
                write_to_eeprom_uint16(0, seed);
                srand(seed);

                // history
                for (int i=0; i<HIST_SIZE; i++){
                    Hist[i] = read_from_eeprom_uint16(HIST_EEPROM_ADDRESS + i*2);
                }

                play_random();
            } else {
                // sequencial play
                Level_0("/");
            }
        }
    }

    while(1)
    {
        wait_ms(500);
        LED = 1 - LED;
    }

}

static uint8_t
Make_SubDir_Path(TCHAR *outpath, uint8_t bufsize, TCHAR *parentdirpath, TCHAR *subdir){
    if ((strlen(parentdirpath) + strlen(subdir)) >= bufsize){
        return 0;
    }
    strcpy(outpath, parentdirpath);
    strcat(outpath, "/");
    strcat(outpath, subdir);
    return 1;
}

/*
 *
 */
static uint8_t
Level_0(char *path){
    F_DIR   dir;
    FILINFO fi;
    FRESULT res1;
    uint8_t mp3count;
    uint8_t dircount;

    // check MP3 file
    count_mp3_dir(path, &mp3count, &dircount);
    if ((mp3count + dircount) == 0) return FR_OK;

    while (1){
        if (mp3count > 0){
            play_subdirectory(path);
        }

        if (dircount > 0){
            res1 = f_opendir(&dir, path);
            if (res1 != FR_OK) break;
            while (1){
                res1 = f_readdir(&dir, &fi);
                if (res1 != FR_OK) break;
                if (fi.fname[0]){
                    if ((fi.fattrib & AM_DIR) != 0){
                        Level_1(path, fi.fname);
                    }
                } else break;
            }
            f_close(&dir);
        }
    }

    return FR_OK;
}

static uint8_t
Level_1(TCHAR *parentdirpath, TCHAR *subdirname){
    TCHAR   path[16];
    F_DIR   dir;
    FILINFO fi;
    FRESULT res1;
    uint8_t mp3count;
    uint8_t dircount;
    uint8_t ret;

    ret = FR_OK;

    if (Make_SubDir_Path(path, 16, parentdirpath, subdirname) == 0){
        return ret;
    }

    // check MP3 file
    count_mp3_dir(path, &mp3count, &dircount);
    if ((mp3count + dircount) == 0) return FR_OK;

    while (1){
        if (mp3count > 0){
            play_subdirectory(path);
        }

        if (dircount > 0){
            res1 = f_opendir(&dir, path);
            if (res1 != FR_OK) break;
            while (1){
                res1 = f_readdir(&dir, &fi);
                if (res1 != FR_OK) break;
                if (fi.fname[0]){
                    if ((fi.fattrib & AM_DIR) != 0){
                        Level_2(path, fi.fname);
                    }
                } else break;
            }
            f_close(&dir);
        }
        
        break;      // MP3_03
    }

    return ret;
}

static uint8_t
Level_2(TCHAR *parentdirpath, TCHAR *subdirname){
    TCHAR   path[16*2];
    F_DIR   dir;
    FILINFO fi;
    FRESULT res1;
    uint8_t mp3count;
    uint8_t dircount;
    uint8_t ret;

    ret = FR_OK;

    if (Make_SubDir_Path(path, 16*2, parentdirpath, subdirname) == 0){
        return ret;
    }

    // check MP3 file
    count_mp3_dir(path, &mp3count, &dircount);
    if ((mp3count + dircount) == 0) return FR_OK;

    while (1){
        if (mp3count > 0){
            play_subdirectory(path);
        }

        if (dircount > 0){
            res1 = f_opendir(&dir, path);
            if (res1 != FR_OK) break;
            while (1){
                res1 = f_readdir(&dir, &fi);
                if (res1 != FR_OK) break;
                if (fi.fname[0]){
                    if ((fi.fattrib & AM_DIR) != 0){
                        Level_3(path, fi.fname);
                    }
                } else break;
            }
            f_close(&dir);
        }
        
        break;      // MP3_03
    }

    return ret;
}

static uint8_t
Level_3(TCHAR *parentdirpath, TCHAR *subdirname){
    TCHAR   path[16*3];
    F_DIR   dir;
    FILINFO fi;
    FRESULT res1;
    uint8_t mp3count;
    uint8_t dircount;
    uint8_t ret;

    ret = FR_OK;

    if (Make_SubDir_Path(path, 16*3, parentdirpath, subdirname) == 0){
        return ret;
    }

    // check MP3 file
    count_mp3_dir(path, &mp3count, &dircount);
    if ((mp3count + dircount) == 0) return FR_OK;

    while (1){
        if (mp3count > 0){
            play_subdirectory(path);
        }

        if (dircount > 0){
            res1 = f_opendir(&dir, path);
            if (res1 != FR_OK) break;
            while (1){
                res1 = f_readdir(&dir, &fi);
                if (res1 != FR_OK) break;
                if (fi.fname[0]){
                    if ((fi.fattrib & AM_DIR) != 0){
//                        Level_4(path, fi.fname);
                    }
                } else break;
            }
            f_close(&dir);
        }
        break;      // MP3_03
    }

    return ret;
}

#if 0
static uint8_t
Level_4(TCHAR *parentdirpath, TCHAR *subdirname){
    TCHAR   path[16*4];
    F_DIR   dir;
    FILINFO fi;
    FRESULT res1;
    uint8_t mp3count;
    uint8_t dircount;
    uint8_t ret;
    uint8_t targetno;
    uint8_t workno;
    uint8_t dirs;

    ret = FR_OK;

    if (Make_SubDir_Path(path, 16*4, parentdirpath, subdirname) == 0){
        return ret;
    }

    // check MP3 file
    count_mp3_dir(path, &mp3count, &dircount);
    //if ((mp3count + dircount) == 0) return FR_OK;
    if (mp3count == 0) return FR_OK;
    dircount = 0;

    while (1){
        //targetno = (rand() % dirs);
        targetno = 0;       // deepst dir
        workno = 0;
        if (mp3count > 0){
            if (TargetAlbum == WorkCount){
            //if (targetno == workno){
                play_subdirectory(path);
            }
            workno++;
            WorkCount++;
            if (FlagEndCountNumAlbum == 0) NumAlbum++;
        }

        if (dircount > 0){
            res1 = f_opendir(&dir, path);
            if (res1 != FR_OK) break;
            while (1){
                res1 = f_readdir(&dir, &fi);
                if (res1 != FR_OK) break;
                if (fi.fname[0]){
                    if ((fi.fattrib & AM_DIR) != 0){
                        //Level_4(path, fi.fname);
                    }
                } else break;
            }
            f_close(&dir);
        }

        break;      // MP3_03
    }

    return ret;
}
#endif


/*
 *  Count mp3 files and sub directories
 *      path : parent directory path
 */
static uint8_t
count_mp3_dir(TCHAR path[], uint8_t *mp3count, uint8_t *dircount){
    F_DIR   dir;
    FILINFO fi;
    FRESULT res1;
    uint8_t mc;
    uint8_t dc;

    mc = dc = 0;
    res1 = f_opendir(&dir, path);
    if (res1 != FR_OK) return res1;
    while (1){
        res1 = f_readdir(&dir, &fi);
        if (res1 != FR_OK) break;
        if (fi.fname[0]){
            if ((fi.fattrib & AM_DIR) == 0 && isMp3(fi.fname)){
                mc++;
            } else if ((fi.fattrib & AM_DIR) != 0){
                dc++;
            }
        } else break;
    }
    f_close(&dir);

    *mp3count = mc;
    *dircount = dc;

    return FR_OK;
}

/*
 *  Play all mp3 files in the directory
 */
static uint8_t
play_subdirectory(TCHAR *path){
    F_DIR   dir;
    FILINFO fi;
    FRESULT res1;
    uint8_t ret;
    int     track_no;

    ret = 0;
    res1 = f_opendir(&dir, path);
    if (res1 != FR_OK){
        return res1;
    }
    track_no = 1;
    for (;;){
        res1 = f_readdir(&dir, &fi);
        if (res1 != FR_OK) break;
        if (fi.fname[0]){       // MP3 name
            if ((fi.fattrib & AM_DIR) == 0 && isMp3(fi.fname)){
                TCHAR   mp3path[64];

                strcpy(mp3path, path);
                strcat(mp3path, "/");
                strcat(mp3path, fi.fname);
                res1 = play_mp3(mp3path);
                if (res1 == 0){
                    // normal end -> next mp3
                } else if (res1 == PLAY_NEXT_MP3){
                } else if (res1 == PLAY_OTHER_DIR){
                    ret = PLAY_OTHER_DIR;
                    break;
                } else {
                    // error
                    ret = res1;
                    break;
                }
            }
        } else {
            break;
        }
    }
    f_close(&dir);

    return ret;
}

static uint16_t
get_numdirs(){
    FRESULT     res;
    FIL         fil;
    TCHAR       buf[64];
    uint16_t    numdirs;

    numdirs = 0;
    res = f_open(&fil, DIRLISTFILE, FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        while (1){
            if (f_gets(buf, 64, &fil) == NULL){
                break;
            }
            numdirs++;
        }
        f_close(&fil);
        return numdirs;
    } else {
        return 0;
    }
}

static uint8_t
play_random(){
    FRESULT     res;
    FIL         fil;
    uint8_t     ret;
    TCHAR       buf[64];
    uint16_t    numdirs, targetno, count;
    char        *c;

    ret = 0;
    numdirs = 0;

    numdirs = get_numdirs();
    if (numdirs == 0) return 0;

    while (1){
        targetno = rand() % numdirs;
        count = 0;
        
        res = f_open(&fil, DIRLISTFILE, FA_READ | FA_OPEN_EXISTING);
        if (res == FR_OK){
            while (1){
                if (f_eof(&fil) == 1){  // EOF OF FILE
                    break;
                }
                if (f_gets(buf, 64, &fil) == NULL){
                    break;
                }
                c = strchr(buf, '\n');
                if (c != NULL) *c = '\0';

                if (count++ == targetno){
                    uint8_t mp3count, dircount;
                    if (count_mp3_dir(buf+2, &mp3count, &dircount) == FR_OK){   // path format-> E:\A\B
                        if (mp3count > 0 && check_hist(targetno) == 0){
                            add_hist(targetno);
                            LED = 1;
                            play_subdirectory(buf+2);
                            LED = 0;
                        }
                    }
                    break;
                }
            }
            f_close(&fil);
        } else {
            return res;
        }
    }

    return ret;
}

/*
 *  check whether folder no exists in history list or not
 */
static uint8_t
check_hist(uint16_t no){
    int     i;

    for (i=0; i<HIST_SIZE; i++){
        if (Hist[i] == no) return 1;
    }
    return 0;
}

/*
 * Add folder no to history list and save to eeprom
 */
static void
add_hist(uint16_t no){
    int     i;

    for (i=0; i<HIST_SIZE; i++){
        if (Hist[i] == 0 || Hist[i] == 0xFFFF){
            Hist[i] = no;
            break;
        };
    }
    if (i == HIST_SIZE){
        i = 0;
        Hist[i] = no;
    }
    write_to_eeprom_uint16(HIST_EEPROM_ADDRESS + i*2, no);
}

/*
 *  Play one MP3 file.
 *      mp3file : Full path name of MP3 file
 *      Return value :  0=current MP3 ended successfully
 *                      PLAY_NEXT_MP3=next MP3 in the current dir
 *                      PLAY_OTHER_DIR
 *                      OTHER=error occured
 */
static uint8_t
play_mp3(TCHAR *mp3file){
    FRESULT     res;
    FIL         fil;
    UINT        br;
    int         idxPlay;
    int         sendbyte;
    int         continue_flag, ret;
    uint16_t    send_buffer_count;
    uint16_t    send_mbyte_count;

    ret = 0;
    res = f_open(&fil, mp3file, FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
        if (res == FR_NO_FILE){
        } else {
        }
        return ret;
    }
    send_buffer_count = 0;
    send_mbyte_count = 0;

    VS10xxRESET();
    idxPlay = 0;
    sendbyte = SEND_SIZE;
    continue_flag = 1;

    while (continue_flag) {
        res = f_read(&fil, MP3DATA, MP3_BUFFER_SIZE, &br);
        if (res != FR_OK || br == 0) break;
        send_buffer_count++;
        if (send_buffer_count == SEND_BUF_CNT_UNIT){
            send_mbyte_count++;
            send_buffer_count = 0;
        }
        idxPlay = 0;

        VS1033SENDDATA(MP3DATA, br);

        if (br < MP3_BUFFER_SIZE){
            continue_flag = 0;
        } else if (STATE_SW1 && CHATTERING_TMR == 0){
            STATE_SW1 = 0;
            if (SW_CONTINUE_COUNT == 0){
                ret = PLAY_NEXT_MP3;
            } else {    // long push
                ret = PLAY_OTHER_DIR;
            }
            SW_CONTINUE_COUNT = 0;
            continue_flag = 0;
        } else if (STATE_SW2 && CHATTERING_TMR == 0){      //
            STATE_SW2 = 0;
            TREBLE_CONTROL++;
            if (TREBLE_CONTROL == 1){
                ST_AMP = 0x09;
                ST_FREQ = 0x03;
                LED_BLINK_MODE = 2;
            /*
            } else if (TREBLE_CONTROL == 2){
                ST_AMP = 0x09;
                ST_FREQ = 0x02;
            } else if (TREBLE_CONTROL == 3){
                ST_AMP = 0x09;
                ST_FREQ = 0x01;
            */
            } else {
                TREBLE_CONTROL = 0;
                ST_AMP = 0x00;
                ST_FREQ = 0x00;
                LED_BLINK_MODE = 1;
            }
            VS10xxCOMM(2, ((ST_AMP << 12) | (ST_FREQ << 8) | (SB_AMP << 4) | SB_FREQ));
        }
    }

    f_close(&fil);
    return ret;
}

static uint8_t
PauseAndSelectItem(){
    uint8_t     ret;

    ret = 0;

    while(1){
        if (STATE_SW1 && CHATTERING_TMR == 0){
            ret = SW1_PRESSED;
            STATE_SW1 = 0;
            break;
        }
        if (STATE_SW2 && CHATTERING_TMR == 0){
            ret = SW2_PRESSED;
            STATE_SW2 = 0;
            break;
        }
        if (SrandSeed == 0xFFFF){
            SrandSeed = 0;
        } else {
            SrandSeed++;
        }
    }

    return ret;
}

static uint8_t
isMp3(char *fname){
    char    *extension;
    extension = strchr(fname, '.');
    if (extension != NULL && strcmp(extension, ".MP3") == 0){
        return 1;
    } else {
        return 0;
    }
}

